/*********************************************************************/
/* file: net.c - do all the net stuff                                */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 

#include <stdio.h>
#ifndef NSTDLIBS
#include <stdlib.h>
#include <unistd.h>
#endif
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OSF1v3
typedef int clockid_t;
#endif
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>

#ifdef SUN
extern "C" {
#endif
#include <arpa/inet.h>
#ifdef SUN
}
#endif

#define TELCMDS 1
#include <arpa/telnet.h>
#undef TELCMDS


#ifndef O_NONBLOCK
#  ifndef FNDELAY
#    define FNDELAY O_NDELAY
#  endif
#  define O_NONBLOCK FNDELAY
#endif


#ifdef NTYPES
  #include "types.h"
#endif

#define TELOPTS 1
#include "structs.h"
#undef TELOPTS
#include "action.h"
#include "llist.h"
#include "main.h"
#include "net.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#ifdef TERM
#  include "term.h"
#endif
#include "ticks.h"
#include "utility.h"
#include "utils.h"

#ifndef BADSIG
#  define BADSIG (void (*)(...))-1
#endif

extern int sessionsstarted, ticker_interrupted, window_interrupted;
extern struct master_table_data master_table;
extern int errno;
extern int signalmask;

#ifdef TERM
#  define CONNECT(sock, addr, len)       term_connect(sock, addr, len)
#  define GETHOSTNAME(name, len)         term_gethostname(name, len)
#  define CLOSE(sock)                    term_close(sock)
#  define GETHOSTBYADDR(addr, len, type) term_gethostbyaddr(addr, len, type)
#  define GETHOSTBYNAME(name)            term_gethostbyname(name)
#else
#  define CONNECT(sock, addr, len)       connect(sock, addr, len)
#  define GETHOSTNAME(name, len)         gethostname(name, len)
#  define CLOSE(sock)                    close(sock)
#  define GETHOSTBYADDR(addr, len, type) gethostbyaddr(addr, len, type)
#  define GETHOSTBYNAME(name)            gethostbyname(name)
#endif

/**************************************************/
/* try connect to the mud specified by the args   */
/* return fd on success / 0 on failure            */
/**************************************************/
#define MAX_HOSTNAME   256

session *cur_con_try;

void break_connect(int i)
{
  cur_con_try->print((const uchar *)"breaking...");
}

void send_errormsg(session *ses)
{
  switch(errno) {
  case ECONNREFUSED:
    ses->print((const uchar *)"CONNECTION REFUSED.\n\r");
    break;
  case ENETUNREACH:
    ses->print((const uchar *)"THE NETWORK IS NOT REACHABLE FROM THIS HOST.\n\r");
    break;
  default:
    ses->print((const uchar *)"Couldn't connect\n\r");
  }
}

int connect_mud(unsigned char *hostname, unsigned char *port,
		struct session *ses)
{
  unsigned char buffer[2048];
  int sock, connectresult, opt = 1, res;
  unsigned long addr;
  struct sockaddr_in sockaddr;
  struct hostent *host;
  fd_set rs, ws, es;
  struct timeval timeout;
#if !defined(HPUX) && !defined(BSD43) && !defined(SUN)
  struct sigaction new_siga, old_siga;
#endif
  
  if(isdigit(*port))  
    sockaddr.sin_port = htons(atoi((const char *)port));
  /* inteprete port part */
  else {
    ses->print((const uchar *)"#THE PORT SHOULD BE A NUMBER.\n\r");
    return 0;
  }

  if(isdigit(*hostname) && (addr = inet_addr((const char *)hostname)) != -1) {
    if(!(host = GETHOSTBYADDR((char *) &addr, sizeof(addr), AF_INET))) {
      sprintf((char *)buffer, "#ERROR - Unknown host %s\n\r", hostname);
      ses->print(buffer);
      return 0;
    }
  } else if ((host = GETHOSTBYNAME((char *)hostname)) == NULL) {
    sprintf((char *)buffer, "#ERROR - Unknown host %s\n\r", hostname);
    ses->print(buffer);
    return 0;
  }

  memcpy((char *)&sockaddr.sin_addr, host->h_addr, sizeof(sockaddr.sin_addr));
  sockaddr.sin_family = AF_INET;

  if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("Couldn't get socket");
    return(0);
  }
  setsockopt(sock, SOL_SOCKET, SO_OOBINLINE, (char *)&opt, sizeof(opt));
  
  if((res = fcntl(sock, F_GETFL, 0)) < 0 ||
     fcntl(sock, F_SETFL, res | O_NONBLOCK) < 0) {
    perror("SETTING NONBLOCK");
    CLOSE(sock);
    return 0;
  }
  
  ses->print((const uchar *)"#Trying to connect...");
  
  sigmask_on();
  connectresult = CONNECT(sock, (struct sockaddr *)&sockaddr,sizeof(sockaddr));
  sigmask_off();
  
  if(connectresult < 0) {
    if(errno == EINPROGRESS) {
      ses->print((const uchar *)"(press ctrl-c to break)...");
      
      FD_ZERO(&rs);
      FD_ZERO(&ws);
      FD_ZERO(&es);
      FD_SET(sock, &rs);
      FD_SET(sock, &ws);
      FD_SET(sock, &es);
      
      timeout.tv_sec = 30;  // Wait 30 secs
      timeout.tv_usec = 0;
      
      cur_con_try = ses;
      
#if !defined(HPUX) && !defined(BSD43) && !defined(SUN)
      bzero((char *)&new_siga, sizeof(new_siga));
      new_siga.sa_flags   = SA_RESTART;
      
      new_siga.sa_handler = break_connect;
      if(sigaction(SIGINT, &new_siga, &old_siga))
	syserr((const uchar *)"signal SIGINT");
#else
      if(signal(SIGINT, break_connect) == BADSIG)
	syserr((const uchar *)"signal SIGINT");
#endif
      
      sigmask_on();
      nocbreak_echo();
      
#ifdef HPUX
      res = select(sock + 1, (int *)&rs, (int *)&ws, (int *)&es, &timeout);
#else
      res = select(sock + 1, &rs, &ws, &es, &timeout);
#endif
      
      cbreak_noecho();
      sigmask_off();
      
#if !defined(HPUX) && !defined(BSD43) && !defined(SUN)
      if(sigaction(SIGINT, &old_siga, 0))
	syserr((const uchar *)"signal SIGINT");
#else
      if(signal(SIGINT, myquitsig) == BADSIG)
	syserr((const uchar *)"signal SIGINT");
#endif

      if(res > 0 && FD_ISSET(sock, &rs)) { //Just check information from socket
	if(recvfrom(sock, buffer, 1, MSG_PEEK, 0, 0) < 0) {
	  send_errormsg(ses);
	  CLOSE(sock);
	  return 0;
	}
      }
      
      if(res <= 0 ||
	 FD_ISSET(sock, &es) ||
	 (!FD_ISSET(sock, &rs) && !FD_ISSET(sock, &ws))) {
	ses->print((const uchar *)"CONNECTION TIMED OUT.\n\r");
	CLOSE(sock);
	return(0);
      }
    } else {
      send_errormsg(ses);
      CLOSE(sock);
      return 0;
    }
  }
  
  ses->print((const uchar *)"Ok\n\r");
  
  sprintf((char *)buffer, "#Connected to %s (%s) port %d.\n\r",
	  inet_ntoa(sockaddr.sin_addr), host->h_name, sockaddr.sin_port);
  ses->print(buffer);

  return sock;
}

/************************************************************/
/* write line to the mud ses is connected to - add \n first */
/************************************************************/
int write_socket(const unsigned char *line, const int net, const int size)
{
  int did_put, sofar;
  
  assert(net >= 0);
  
  sofar = 0;
  do {    /* write until entire string has been written */
    did_put = write(net, line + sofar, size - sofar);
    if(did_put < 0)
      return(-1);
    
    sofar += did_put;
  } while( sofar < size);
  
  return(sofar);
}

session *session::write_net(const unsigned char *line)
{
  unsigned char outtext[BUFFER_SIZE + 1];
  int size;
  
  size = str_cat(outtext, 0, sizeof(outtext), line);
  size = str_cat(outtext, size, sizeof(outtext), (const uchar *)"\n");
  
  if(write_socket(outtext, net, size) < 0) {
    perror("write socket");
    delete this;
    return(0);
  }
  
  if(master_table.ucommand) {
    *promptbuf = '\0';
    prompt_pos = 0;
  }
  return(this);
}

session *session::send_naws()
{
  unsigned char tmp[20];
  int pos, lines;
  
  if(net >= 0 && will_status[TELOPT_NAWS] == WILL) {
    lines = output->get_size()->end - output->get_size()->start + 1;
    if(output->get_time_on())
      lines--;
    
    pos = 0;
    tmp[pos++] = IAC;
    tmp[pos++] = SB;
    tmp[pos++] = TELOPT_NAWS;
    
#define PUTSHORT(x) { if((tmp[pos++] = ((x)>>8)&0xff) == IAC) \
                        tmp[pos++] = IAC; \
			  if((tmp[pos++] = ((x))&0xff) == IAC) \
			    tmp[pos++] = IAC; }
    PUTSHORT(window::get_cols());
    PUTSHORT(lines);
    tmp[pos++] = IAC;
    tmp[pos++] = SE;
    
    if(write_socket(tmp, net, pos) < 0) {
      perror("write socket");
      delete this;
      return(0);
    }
  }
  return(this);
}
#undef PUTSHORT
 
// ***************************************
//  check actions and process substitutes
// ***************************************

session *process_text(const uchar *line, uchar *dest, int &len, session *ses)
{
  uchar buffer[BUFFER_SIZE], linebuf[BUFFER_SIZE], buf[BUFFER_SIZE];
  int lastsub = TRUE;
  struct listnode *ln;
  
  len = str_cat(dest, 0, BUFFER_SIZE, line);
  
  if(IS_SET(ses->ignore_flags, IGNORE_ACTION) &&
     IS_SET(ses->ignore_flags, IGNORE_SUBS))
    return(ses);
  
  for(ln = ses->actions; ln; ln = ln->next) {
    if((ln->type == TYPE_ACTION && IS_SET(ses->ignore_flags, IGNORE_ACTION)) ||
       ((ln->type == TYPE_SUBS || ln->type == TYPE_GAGS) &&
	IS_SET(ses->ignore_flags, IGNORE_SUBS)))
      continue;
    
    substitute_vars(ln->left, buffer, sizeof(buffer), FALSE, ses);
    if(ln->type == TYPE_ACTION) {
      if(lastsub) {
	strip_ctrl_codes(linebuf, dest);
	lastsub = FALSE;
      }
      if(check_one_action(linebuf, buffer, TRUE)) {
	substitute_vars(ln->right, buffer, sizeof(buffer), TRUE, ses, FALSE);
	if(IS_SET(ses->echo_flags, ECHO_ACTION)) {
	  sprintf((char *)buf, "[ACTION: %s]\n\r", buffer);
	  ses->print(buf);
	}
	if(!parse_input(buffer, ses))
	  return(0);
      }
    } else {
      if(check_one_action(dest, buffer)) {
	lastsub = TRUE;
	
	if(ln->type == TYPE_GAGS) {
	  *dest = '\0';
	  len = 0;
	  return(ses);
	} else
	  len = substitute_vars(ln->right, dest, BUFFER_SIZE, TRUE, ses);
      }
    }
  }
  return(ses);
}

/*******************************************************************/
/* read at most BUFFER_SIZE chars from mud - parse protecol stuff  */
/*******************************************************************/
void session::read_net()
{
  unsigned char readbuffer[512], endbuffer[BUFFER_SIZE];
  unsigned char linebuffer[BUFFER_SIZE], subresult[BUFFER_SIZE];
  unsigned char *rbuf, *dest, *str, *tmp, letter, *endstr;
  int len, pos, spos, middle, i, cutline;
  struct timeval null_time;
  fd_set input_set, exc_set;
  
  null_time.tv_sec = 0;
  null_time.tv_usec = 0;
  
  dest = linebuffer;
  rbuf = readbuffer;
  if(last_text.len) {
    middle = FALSE;
    str = last_text.text;
    while(str < last_text.text + last_text.len) {
      switch(*str) {
      case ESC:
	tmp = str;
	
	str++;
	middle = TRUE;
	if(*str == '[') {
	  str++;
	  if(*str == '?')
	    str++;
	  while(isdigit(*str))
	    str++;
	  while(*str == ';') {
	    str++;
	    while(isdigit(*str))
	      str++;
	  }
	  if(*str) {
	    str++;
	    middle = FALSE;
	  }
	} else {
	  if(*str) {
	    str++;
	    middle = FALSE;
	  }
	}
	if(middle) {
	  while(*tmp)
	    *rbuf++ = *tmp++;
	} else {
	  while(tmp < str)
	    *dest++ = *tmp++;
	}
	break;
      case IAC:
	while(str < last_text.text + last_text.len)
	  *rbuf++ = *str++;
	middle = TRUE;
	break;
      default:
	*dest++ = *str++;
	break;
      }
    }
    free(last_text.text);
    last_text.text = 0;
    last_text.len  = 0;
  }
  *rbuf = '\0';
  *dest = '\0';
  assert(rbuf - readbuffer < sizeof(readbuffer));
  
  do {
    middle = FALSE;
    switch((len = read(net, rbuf,
		       sizeof(readbuffer) - (rbuf - readbuffer) - 1))) {
    case 0:
      print((const uchar *)"#Connection closed by foreing host.\n\r");
      delete this;
      prompt();
      return;
      break;
    case -1:
#ifdef DEBUG
      if(errno == EWOULDBLOCK) {
	print((const uchar *)"###INFO: got EWOULDBLOCK\n\r");
      } else {
	perror("#Read from socket descriptor failded.");
	delete this;
	prompt();
      }
#else
      perror("#Read from socket descriptor failded.");
      delete this;
      prompt();
#endif
      return;
    default:
      *(rbuf + len) = '\0';
      len += rbuf - readbuffer;
      rbuf = readbuffer;
      
      spos = 0;
      while(spos < len) {
	cutline = FALSE;
	endstr  = endbuffer;
	
	switch(readbuffer[spos]) {
	case '\0':
	  spos++;
	  break;
	case ESC:
	  i = spos + 1;
	  
	  middle = TRUE;
	  if(readbuffer[i] == '[') {
	    i++;
	    if(readbuffer[i] == '?')
	      i++;
	    while(isdigit(readbuffer[i]))
	      i++;
	    while(readbuffer[i] == ';') {
	      i++;
	      while(isdigit(readbuffer[i]))
		i++;
	    }
	    if((letter = readbuffer[i])) {
	      i++;
	      middle = FALSE;
	      if(letter == 'H' || letter == 'A' || letter == 'B')
		cutline = TRUE;
	    }
	  } else {
	    if((letter = readbuffer[i])) {
	      i++;
	      middle = FALSE;
	      if(letter == 'D' || letter == 'M')
		cutline = TRUE;
	    }
	  }
	  if(middle) {
	    while(spos < len)
	      *rbuf++ = readbuffer[spos++];
	    *rbuf = '\0';
	  } else {
	    if(cutline) {
	      while(spos < i)
		*endstr++ = readbuffer[spos++];
	    } else {
	      while(spos < i)
		*dest++ = readbuffer[spos++];
	    }
	  }
	  break;
	case IAC:
	  i = spos + 1;
	  middle = TRUE;
	  if(readbuffer[i] == SB) {
	    i++;
	    while(readbuffer[i] != IAC && readbuffer[i + 1] != SE)
	      i++;
	    if(readbuffer[i] == IAC && readbuffer[i + 1] == SE) {
	      middle = FALSE;
	      switch(readbuffer[spos + 2]) {
	      case TELOPT_TTYPE:
		switch(readbuffer[spos + 3]) {
		case TELQUAL_IS:
		  break;
		case TELQUAL_SEND:
		  
		  subresult[0] = IAC;
		  subresult[1] = SB;
		  subresult[2] = TELOPT_TTYPE;
		  subresult[3] = TELQUAL_IS;
		  if((tmp = (uchar *)getenv("TERM"))) {
		    pos = str_cat(subresult, 4, sizeof(subresult), tmp);
		  } else {
		    pos = str_cat(subresult, 4, sizeof(subresult),
				  DEF_TERM_TYPE);
		  }
		  subresult[pos++] = IAC;
		  subresult[pos++] = SE;
		  if(write_socket(subresult, net, pos) < 0) {
		    perror("write socket");
		    delete this;
		    return;
		  }
		  break;
		}
		break;
	      case TELQUAL_INFO:
		break;
	      case TELQUAL_NAME:
		break;
	      } /* switch */
	      spos = i + 2;
	    }
	  } else {
	    if(i + 1 < len) {
	      middle = FALSE;
	      if(!do_telnet_protecol(readbuffer[i], readbuffer[i + 1]))
		return;
	      
	      spos += 3;
	    }
	  }
	  if(middle) {
	    while(spos < len)
	      *rbuf++ = readbuffer[spos++];
	    *rbuf = '\0';
	  }
	  break;
	case '\f':
	  cutline = TRUE;
	  while(readbuffer[spos] == '\f') {
	    *endstr++ = '\f';
	    spos++;
	  }
	  break;
	case '\n':
	  cutline   = TRUE;
	  *endstr++ = '\n';
	  spos++;
	  if(readbuffer[spos] == '\r') {
	    *endstr++ = '\r';
	    while(readbuffer[spos] == '\r')
	      spos++;
	  } else
	    *endstr++ = '\r';
	  break;
	case '\r':
	  cutline   = TRUE;
	  *endstr++ = '\r';
	  while(readbuffer[spos] == '\r')
	    spos++;
	  break;
	default:
	  *dest++ = readbuffer[spos++];
	  break;
	}

	if(cutline) {
	  *dest = '\0';
	  *endstr = '\0';
	  
	  if(logfile)
	    fprintf(logfile, "%s\n", linebuffer);
	  
	  if(!process_text(linebuffer, subresult, pos, this))
	    return;

	  if(*subresult || !*linebuffer) {
	    pos = str_cat(subresult, pos, sizeof(subresult), endbuffer);
	  }
	  print(subresult);
	  dest = linebuffer;
	}
      }
      *dest = '\0';
    } /* switch */
    
    FD_ZERO(&input_set);
    FD_ZERO(&exc_set);
    FD_SET(net, &input_set);
    FD_SET(net, &exc_set);
    
    sigsetmask(signalmask);
#ifdef HPUX
    if(select(net + 1, (int *)&input_set, (int *)0, (int *)&exc_set,
	      &null_time) < 0)
      syserr((const uchar *)"net.cc select");
#else
    if(select(net + 1, &input_set, (fd_set *)0, &exc_set, &null_time) < 0)
      syserr((const uchar *)"select");
#endif
    sigsetmask(0);
  } while(FD_ISSET(net, &input_set));
  
  *dest = '\0';
  *rbuf = '\0';
  if(middle) {
    len = rbuf - readbuffer + dest - linebuffer;
    CREATE(last_text.text, uchar, len + 1);
    pos = str_cat(last_text.text, 0, len + 1, linebuffer);
    pos = str_cat(last_text.text, pos, len + 1, readbuffer);
    last_text.len = pos;
    *promptbuf = '\0';
    prompt_pos = 0;
  } else if(*linebuffer) {
    if(logfile)
      fprintf(logfile, "%s", linebuffer);
    
    if(!process_text(linebuffer, subresult, pos, this))
      return;

    print(subresult);
    if(output != input &&
       input->get_active_ses() == this &&
       !input->get_hist_on())
      input->print(subresult);
    set_prompt(subresult);
  } else {
    *promptbuf = '\0';
    prompt_pos = 0;
  }
  prompt();
}

/*****************************************************************/
/* respond according to the telnet protecol - weeeeelllllll..... */
/*****************************************************************/

session *session::do_telnet_protecol(int telcom, int telopt)
{
  /* we don't do anything here.. why should we? add the stuff yourself if
     you feel like being nice..... */
  
  unsigned char answer[20];
  int found = FALSE, needanswer = FALSE, size, lines;
  
  if(telopt < 0 || telopt >= NTELOPTS)
    return(this);
  
  answer[0] = IAC;
  answer[2] = telopt;
  answer[3] = '\0';
  
  switch(telcom) {
  case DONT:
    switch(telopt) {
    case TELOPT_ECHO:
      if(will_status[telopt] != WONT) {
	needanswer = TRUE;
	line_mode = FALSE;
	will_status[telopt] = WONT;
      }
      found = TRUE;
      break;
    default:
      if(will_status[telopt] != WONT) {
	needanswer = TRUE;
	will_status[telopt] = WONT;
      }
      found = TRUE;
      break;
    }
    if(needanswer) {
      answer[1] = will_status[telopt];
      if(write_socket(answer, get_socket(), 3) < 0) {
	perror("write socket");
	delete this;
	return(0);
      }
    }
    break;
  case DO:
    size = 3;
    switch(telopt) {
    case TELOPT_ECHO:
      if(will_status[telopt] != WILL) {
	needanswer = TRUE;
	line_mode = TRUE;
	will_status[telopt] = WILL;
      }
      found = TRUE;
      break;
    case TELOPT_STATUS:
      if(will_status[telopt] != WONT) {   // WILL
	needanswer = TRUE;
	will_status[telopt] = WONT;
      }
      found = TRUE;
      break;
    case TELOPT_TTYPE:
      if(will_status[telopt] != WILL) {
	needanswer = TRUE;
	will_status[telopt] = WILL;
      }
      found = TRUE;
      break;
    case TELOPT_NAWS:
      if(will_status[telopt] != WILL) {
	needanswer = TRUE;
	will_status[telopt] = WILL;
	
	lines = output->get_size()->end - output->get_size()->start + 1;
	if(output->get_time_on())
	  lines--;
	
	answer[size++] = IAC;
	answer[size++] = SB;
	answer[size++] = TELOPT_NAWS;
	
#define PUTSHORT(x) { if((answer[size++] = ((x)>>8)&0xff) == IAC) \
			answer[size++] = IAC; \
			  if((answer[size++] = ((x))&0xff) == IAC) \
			    answer[size++] = IAC; }
	PUTSHORT(window::get_cols());
	PUTSHORT(lines);
#undef PUTSHORT
	answer[size++] = IAC;
	answer[size++] = SE;
      }
      found = TRUE;
      break;
    default:
      if(will_status[telopt] != WONT) {
	needanswer = TRUE;
	will_status[telopt] = WONT;
      }
      break;
    }
    if(needanswer) {
      answer[1] = will_status[telopt];
      if(write_socket(answer, get_socket(), size) < 0) {
	perror("write socket");
	delete this;
	return(0);
      }
      if(answer[3] != '\0')
	answer[3] = '\0';
    }
    break;
  case WONT:
    switch(telopt) {
    case TELOPT_ECHO:
      if(do_status[telopt] != DONT) {
	term_echoing = TRUE;
	do_status[telopt] = DONT;
      }
      found = TRUE;
      break;
    default:
      if(do_status[telopt] != DONT) {
	do_status[telopt] = DONT;
      }
      break;
    }
    answer[1] = do_status[telopt];
    if(write_socket(answer, get_socket(), 3) < 0) {
      perror("write socket");
      delete this;
      return(0);
    }
    break;
  case WILL:
    switch(telopt) {
    case TELOPT_ECHO:
      if(do_status[telopt] != DO) {
	term_echoing = FALSE;
	window::get_active_win()->remove_user_text();
	do_status[telopt] = DO;
      }
      found = TRUE;
      break;
    case TELOPT_STATUS:
      if(do_status[telopt] != DONT) {
	do_status[telopt] = DONT;
      }
      found = TRUE;
      break;
    case TELOPT_SGA:
      if(do_status[telopt] != DO) {
	do_status[telopt] = DO;
      }
      found = TRUE;
      break;
    default:
      if(do_status[telopt] != DONT) {
	do_status[telopt] = DONT;
      }
      break;
    }
    answer[1] = do_status[telopt];
    if(write_socket(answer, get_socket(), 3) < 0) {
      perror("write socket");
      delete this;
      return(0);
    }
    break;
  } /* switch */
  
#ifdef DEBUG
  if(!found) {
    unsigned char buf[256];
    
#if defined(TELCMD_OK) && defined(TELCMD)
    sprintf((char *)buf,
	    "#UNSUPORTED TELNET OPTION: com:%d(%s) opt:%d(%s)\n\r",
	    telcom,
	    TELCMD_OK(telcom) ? TELCMD(telcom) : "unknown",
	    telopt,
	    TELOPT_OK(telopt) ? TELOPT(telopt) : "unknown");
#else
    sprintf((char *)buf, "#UNSUPORTED TELNET OPTION: com:%d opt:%d\n\r",
	    telcom,
	    telopt,
	    TELOPT_OK(telopt) ? TELOPT(telopt) : "unknown");
#endif
    print(buf);
    print((const uchar *)"#If you got this message with some dikumud email to otulonen@snakemail.hut.fi\n\r");
    print((const uchar *)"#those com and opt numbers. (please)\n\r");
  }
#endif
  return(this);
}
